unit Math3D;

{

 3D Maths (c)2002 by Paul TOTH <tothpaul@free.fr>
 http://tothpaul.free.fr

 based on a lot of free stuff all over the net :)
 thank you very much !
}

{
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
}

interface

uses
 Math;

type
 TVector3f=record
  x,y,z:single;
 end;

 TVector4f=record
  x,y,z,w:single;
 end;

 TMatrix16f=array[0..3,0..3] of single;

function AddVector(const a,b:TVector3f):TVector3f;
function SubVector(const a,b:TVector3f):TVector3f;

function ScaleVector(const a:TVector3f; scale:single):TVector3f;

function VectorLength(const a:TVector3f):single;

procedure NormalizeVector(var a:TVector3f);

function DotProduct(const a,b:TVector3f):single;
function CrossProduct(const a,b:TVector3f):TVector3f;

function FaceNormal(const a,b,c:TVector3f):TVector3f;

procedure IdentityMatrix(Var Matrix:TMatrix16f);
function MulMatrix(const A,B:TMatrix16f):TMatrix16f;
function InvertMatrix(Const Matrix:TMatrix16f; var Invert:TMatrix16f):boolean;
function ApplyMatrix(Const Matrix:TMatrix16f; const Vector:TVector3f):TVector3f;

implementation

function AddVector(const a,b:TVector3f):TVector3f;
 begin
  Result.x:=a.x+b.x;
  Result.y:=a.y+b.y;
  Result.z:=a.z+b.z;
 end;

function SubVector(const a,b:TVector3f):TVector3f;
 begin
  Result.x:=a.x-b.x;
  Result.y:=a.y-b.y;
  Result.z:=a.z-b.z;
 end;

function ScaleVector(const a:TVector3f; scale:single):TVector3f;
 begin
  Result.x:=scale*a.x;
  Result.y:=scale*a.y;
  Result.z:=scale*a.z;
 end;

function VectorLength(const a:TVector3f):single;
 begin
  Result:=sqrt((a.x*a.x) + (a.y*a.y) + (a.z*a.z));
 end;

function DotProduct(const a,b:TVector3f):single;
 begin
  result:= (a.z * b.z) + (a.x * b.x) + (a.y * b.y);
 end;

function CrossProduct(const a,b:TVector3f):TVector3f;
 begin
  Result.x:=a.y * b.z - a.z * b.y;
  Result.y:=a.z * b.x - a.x * b.z;
  Result.z:=a.x * b.y - a.y * b.x;
 end;

procedure NormalizeVector(var a:TVector3f);
 var
  l:single;
 begin
  l:=VectorLength(a);
  if l<>0 then l:=1/l;
  a.x:=a.x*l;
  a.y:=a.y*l;
  a.z:=a.z*l;
 end;

function FaceNormal(const a,b,c:TVector3f):TVector3f;
 var
  u,v:TVector3f;
  theta:single;
 begin
  u:=SubVector(b,a);
  v:=SubVector(c,a);
  NormalizeVector(u);
  NormalizeVector(v);
  result:=ScaleVector(CrossProduct(u, v),1/sin(ArcCos(DotProduct(u,v))));
 end;

procedure IdentityMatrix(Var Matrix:TMatrix16f);
 begin
  FillChar(Matrix,SizeOf(Matrix),0);
  Matrix[0,0]:=1;
  Matrix[1,1]:=1;
  Matrix[2,2]:=1;
  Matrix[3,3]:=1;
 end;

function MulMatrix(const A,B:TMatrix16f):TMatrix16f;
 var
  i,j:integer;
 begin
  for i:=0 to 3 do
   for j:=0 to 3 do
    Result[i,j]:=a[i,0]*b[0,j]
                +a[i,1]*b[1,j]
                +a[i,2]*b[2,j]
                +a[i,3]*b[3,j];
 end;

function InvertMatrix(Const Matrix:TMatrix16f; var Invert:TMatrix16f):boolean;
// from SGI GLU source code
 var
  temp:TMatrix16f;
  i,j,k,swap:integer;
  t:single;
 begin
  temp:=Matrix;
  IdentityMatrix(Invert);
  for i:=0 to 3 do begin
   // look for largest element in colum
   swap:=i;
   for j:=i+1 to 3 do
    if abs(temp[j,i])>abs(temp[i,i]) then swap:=j;
   if swap<>i then begin
    // swap rows
    for k:=0 to 3 do begin
     t:=temp[i,k];
     temp[i,k]:=temp[swap,k];
     temp[swap,k]:=t;
     t:=invert[i,k];
     invert[i,k]:=invert[swap,k];
     invert[swap,k]:=t;
    end;
   end;
   t:=temp[i,i];
   if t=0 then begin
   // no non-zero pivot. The matrix is singular, which shouldn't
   // happen. This means the user gave us a bad matrix.
    result:=false;
    exit;
   end;
   for k:=0 to 3 do begin
    temp[i,k]:=temp[i,k]/t;
    invert[i,k]:=invert[i,k]/t;
   end;
   for j:=0 to 3 do begin
    if j<>i then begin
     t:=temp[j,i];
     for k:=0 to 3 do begin
      temp[j,k]:=temp[j,k]-temp[i,k]*t;
      invert[j,k]:=invert[j,k]-invert[i,k]*t;
     end;
    end;
   end;
  end;
  Result:=True;
 end;

function ApplyMatrix(Const Matrix:TMatrix16f; Const Vector:TVector3f):TVector3f;
 var
  w:single;
 begin
  result.x:=Matrix[0,0]*Vector.x+Matrix[1,0]*Vector.y+Matrix[2,0]*Vector.z+Matrix[3,0];
  result.y:=Matrix[0,1]*Vector.x+Matrix[1,1]*Vector.y+Matrix[2,1]*Vector.z+Matrix[3,1];
  result.z:=Matrix[0,2]*Vector.x+Matrix[1,2]*Vector.y+Matrix[2,2]*Vector.z+Matrix[3,2];
         w:=Matrix[0,3]*Vector.x+Matrix[1,3]*Vector.y+Matrix[2,3]*Vector.z+Matrix[3,3];
  if w=0 then exit;
  result.x:=result.x/w;
  result.y:=result.y/w;
  result.z:=result.z/w;
 end;

end.
